#include "x86struct.h"
#include "x86.h"
#include "x86mem.h"
#include "lib.h"


/*****************************************************************************
 * console.c
 *
 *   This file contains code for clearing and writing to the console.
 *
 *   YOU DO NOT NEED TO UNDERSTAND THIS!
 *
 *****************************************************************************/


/*****************************************************************************
 * console_clear
 *
 *   Clear the console by writing spaces to it, and move the cursor to the
 *   lower right (row 24, column 79).
 *
 *****************************************************************************/

void
console_clear(void)
{
	int i;
	uint16_t *consolebuf = (uint16_t *) CONSOLE_BEGIN; 
	
	for (i = 0; i < 80 * 25; i++)
		consolebuf[i] = ' ' | 0x0700;
	outb(0x3D4, 14);
	outb(0x3D5, (24*80 + 79) / 256);
	outb(0x3D4, 15);
	outb(0x3D5, (24*80 + 79) % 256);
}


/*****************************************************************************
 * console_vprintf
 *
 *   Print a message onto the console, starting at the given row and
 *   column.  Supports a limited set of printf-type formatting commands,
 *   including "%d" for decimal integers and "%x" for hexidecimal numbers.
 *
 *****************************************************************************/

static uint16_t console_color = 0x0700;

static uint16_t *
console_putc(uint16_t *cptr, unsigned char c, int color)
{
	if (cptr >= CONSOLE_END)
		/* do nothing */;
	else if (c == '\n') {
		int pos = (cptr - CONSOLE_BEGIN) % 80;
		for (; pos != 80; pos++)
			*cptr++ = ' ' | color;
	} else
		*cptr++ = c | color;
	return cptr;
}

static const char upper_digits[] = "0123456789ABCDEF";
static const char lower_digits[] = "0123456789abcdef";

static char *
fill_numbuf(char *numbuf_end, uint32_t val, int base, const char *digits)
{
	*--numbuf_end = '\0';
	do {
		*--numbuf_end = digits[val % base];
		val /= base;
	} while (val != 0);
	return numbuf_end;
}

void
console_vprintf(int row, int col, const char *format, va_list val)
{
	uint16_t *cptr = CONSOLE_BEGIN + 80 * row + col;
	int color = console_color;
	int space, width, negative, len, altflag;
#define NUMBUFSIZ 20
	char numbuf[NUMBUFSIZ];
	char *data;

	for (; *format && cptr < CONSOLE_END; format++) {
		if (*format != '%') {
			cptr = console_putc(cptr, *format, color);
			continue;
		}

		negative = 0;
		space = ' ';
		width = -1;
		altflag = 0;
	reswitch:
		format++;
		switch (*format) {
		case '0':
			space = '0';
			goto reswitch;
		case '1': case '2': case '3': case '4':
		case '5': case '6': case '7': case '8': case '9':
			if (width != -1)
				goto normal;
			for (width = 0; *format >= '0' && *format <= '9'; format++)
				width = 10 * width + *format - '0';
			format--;
			goto reswitch;
		case '#':
			altflag = 1;
			goto reswitch;
		case 'd': {
			int v = va_arg(val, int);
			data = fill_numbuf(numbuf + NUMBUFSIZ, v > 0 ? v : -v, 10, upper_digits);
			if (v < 0)
				negative = 1;
			goto print;
		}
		case 'u': {
			unsigned v = va_arg(val, unsigned);
			data = fill_numbuf(numbuf + NUMBUFSIZ, v, 10, upper_digits);
			goto print;
		}
		case 'x': {
			unsigned v = va_arg(val, unsigned);
			data = fill_numbuf(numbuf + NUMBUFSIZ, v, 16, lower_digits);
			goto print;
		}
		case 'X': {
			unsigned v = va_arg(val, unsigned);
			data = fill_numbuf(numbuf + NUMBUFSIZ, v, 16, upper_digits);
			goto print;
		}
		case 's':
			data = va_arg(val, char *);
			goto print;
		case 'C':
			color = va_arg(val, int);
			if (altflag)
				console_color = color;
			goto done;
		case 'c':
			data = numbuf;
			numbuf[0] = va_arg(val, int);
			numbuf[1] = '\0';
			goto print;
		normal:
		default:
			data = numbuf;
			numbuf[0] = (*format ? *format : '%');
			numbuf[1] = '\0';
			if (!*format)
				format--;
			goto print;
		}
	print:
		if (negative)
			cptr = console_putc(cptr, '-', color), width--;
		for (len = strlen(data); len < width; width--)
			cptr = console_putc(cptr, space, color);
		for (; *data; data++)
			cptr = console_putc(cptr, *data, color);
	done: ;
	}
}


/*****************************************************************************
 * console_printf
 *
 *   Calls console_vprintf with the arguments taken from its argument list.
 *
 *****************************************************************************/

void
console_printf(int row, int col, const char *format, ...)
{
	va_list val;
	va_start(val, format);
	console_vprintf(row, col, format, val);
	va_end(val);
}


/*****************************************************************************
 * panic
 *
 *   Print an error message to the lower part of the console,
 *   then spin forever.
 *
 *****************************************************************************/

void
panic(const char *message, ...)
{
	va_list val;
	va_start(val, message);
	console_printf(23, 0, "%#CPANIC: ", 0xC000);
	console_vprintf(23, 7, message, val);
	va_end(val);

	while (1)
		/* do nothing */;
}
