#include "x86.h"
#include "elf.h"
#include "memos-kern.h"
#include "x86struct.h"
#include "x86mem.h"

/*****************************************************************************
 * memos-loader.c
 *
 *   Load memos processes in from disk.
 *   Uses page_alloc() to allocate memory for each process.
 *
 *   Don't worry about understanding this loader!
 *
 *****************************************************************************/

#define SECTORSIZE		512
#define SECTORS_PER_PROCESS	32

// a free page for the ELF header
static char elf_header_buffer[PAGESIZE] __attribute__ ((aligned (PAGESIZE)));

static void readsect(void *addr, uint32_t sectornum);
static void readseg(process_t *process, uint32_t va, uint32_t filesz, uint32_t memsz, uint32_t sectorno);

void
process_load(process_t *process)
{
	struct Proghdr *ph, *eph;
	int sector_offset = 1 + (process->p_pid + 1) * SECTORS_PER_PROCESS;
	struct Elf *elf_header = (struct Elf *) elf_header_buffer;

	// read 1st page off disk
	readseg(NULL, (uint32_t) elf_header_buffer, PAGESIZE, PAGESIZE,
		sector_offset);

	// is this a valid ELF?
	if (elf_header->e_magic != ELF_MAGIC)
		return;

	// load each program segment (ignores ph flags)
	ph = (struct Proghdr*) (elf_header_buffer + elf_header->e_phoff);
	eph = ph + elf_header->e_phnum;
	for (; ph < eph; ph++)
		if (ph->p_type == ELF_PROG_LOAD)
			readseg(process, ph->p_va, ph->p_filesz, ph->p_memsz,
				sector_offset + ph->p_offset / SECTORSIZE);

	// store the entry point from the ELF header into the stack
	process->p_registers.reg_eip = elf_header->e_entry;
}

// Read 'filesz' bytes at 'offset' from kernel into virtual address 'va'.
// Might copy more than asked
void
readseg(process_t *proc, uint32_t va, uint32_t filesz, uint32_t memsz,
	uint32_t sectorno)
{
	uint32_t end_va;
	uint32_t end_page;

	end_va = va + filesz;
	memsz += va;

	// round down to sector boundary
	va &= ~(SECTORSIZE - 1);
	end_page = va & ~(PAGESIZE - 1);

	// read sectors
	while (va < end_va) {
		if (va >= end_page && proc) {
			page_alloc(proc, end_page);
			end_page += PAGESIZE;
		}
		readsect((uint8_t*) va, sectorno);
		va += SECTORSIZE;
		sectorno++;
	}

	// clear bss segment
	while (end_va < memsz) {
		if (end_va >= end_page && proc) {
			page_alloc(proc, end_page);
			end_page += PAGESIZE;
		}
		*((uint8_t*) end_va++) = 0;
	}
}

static void
waitdisk(void)
{
	// wait for disk reaady
	while ((inb(0x1F7) & 0xC0) != 0x40)
		/* do nothing */;
}

static void
readsect(void *dst, uint32_t offset)
{
	// wait for disk to be ready
	waitdisk();

	outb(0x1F2, 1);		// count = 1
	outb(0x1F3, offset);
	outb(0x1F4, offset >> 8);
	outb(0x1F5, offset >> 16);
	outb(0x1F6, (offset >> 24) | 0xE0);
	outb(0x1F7, 0x20);	// cmd 0x20 - read sectors

	// wait for disk to be ready
	waitdisk();

	// read a sector
	insl(0x1F0, dst, SECTORSIZE/4);
}
