#include "memos-kern.h"
#include "x86struct.h"
#include "x86.h"
#include "x86mem.h"


/*****************************************************************************
 * memos-pages.c
 *
 *   Managing physical and virtual memory.
 *
 *****************************************************************************/


/*****************************************************************************
 * PAGE LABELS
 *
 *   The page_labels[] array keeps track of a *label* for each physical page.
 *   'page_labels[pn]' equals the label of physical page number 'pn'.
 *   The label indicates whether a page has been used, and if so, what for.
 *   The 'show_physical_memory' and 'show_virtual_memory' functions examine
 *   page labels to decide what character to print for each page.
 *   Possible label values are:
 *
 *   Label                Character  Use
 *   -----                ---------  ---
 *   MEMLABEL_FREE == 0       .      Free
 *   MEMLABEL_RESERVED        R      Reserved (e.g. I/O memory)
 *   MEMLABEL_KERNEL          K      Kernel code and/or data
 *   MEMLABEL_PAGETABLE       P      Page directory or page table
 *   MEMLABEL_PROC + 0        1      Process PID 0 code and/or data
 *   MEMLABEL_PROC + 1        2      Process PID 1 code and/or data
 *   MEMLABEL_PROC + 2        3      Process PID 2 code and/or data
 *   MEMLABEL_PROC + 3        4      Process PID 3 code and/or data
 *
 *   The memory_labels_init() function sets up page labels for the kernel's
 *   initial state: I/O memory is MEMLABEL_RESERVED, kernel memory is
 *   MEMLABEL_KERNEL, and everything else is MEMLABEL_FREE.
 *
 *****************************************************************************/

#define MEMLABEL_FREE		0
#define MEMLABEL_RESERVED	1
#define MEMLABEL_PAGETABLE	2
#define MEMLABEL_KERNEL		3
#define MEMLABEL_PROC		4

#define NPAGES			(MAX_PHYSICAL_MEMORY / PAGESIZE)
static int page_labels[NPAGES];


/*****************************************************************************
 * page_alloc
 *
 *   Allocate a page of memory for process 'proc' at address 'address'.
 *   Return -1 if no page is available.
 *
 *   The initial code allocates the physical page at physical address
 *   'address'.
 *   As the minilab progresses, you will change this function to *map* the
 *   chosen page at *virtual* address 'address', and then to allocate
 *   *any* free physical page (instead of the physical page at 'address')!
 *   You will find the page_get_free and page_map_user functions useful.
 *
 *****************************************************************************/

static int page_map_user(process_t *proc, int physical_page_number,
			 uint32_t address);
	// Map the physical page numbered 'physical_page_number'
	// at virtual address 'address' in 'proc's address space.

static int page_get_free(void);
	// Return the physical page number of an unallocated page
	// (a page with page_label MEMLABEL_FREE).

int
page_alloc(process_t *proc, uint32_t address)
{
	int physical_page_number;

	// Physical address allocation:
	// Use the physical page corresponding to the desired address
	physical_page_number = PAGENUM(address);
	if (physical_page_number >= NPAGES)
		panic("Attempt to allocate nonexistent physical page %d!",
		      physical_page_number);

	// Label the physical page as belonging to this process
	if (page_labels[physical_page_number] != MEMLABEL_FREE)
		panic("Attempt to reallocate physical page %d!",
		      physical_page_number);
	page_labels[physical_page_number] = MEMLABEL_PROC + proc->p_pid;

	return 0;
}


/*****************************************************************************
              YOU DO NOT NEED TO CHANGE ANY CODE BELOW THIS POINT
 
 *****************************************************************************/

/*****************************************************************************
 * page_get_free
 *
 *   Return the physical page number of a free memory page,
 *   or -1 if all of physical memory has been allocated.
 *
 *****************************************************************************/

static int
page_get_free(void)
{
	int pn;
	for (pn = 0; pn < NPAGES; pn++)
		if (page_labels[pn] == MEMLABEL_FREE)
			return pn;
	console_printf(24, 0, "%COut of physical memory!", 0x0C00);
	return -1;
}


/*****************************************************************************
 * page_map_user
 *
 *   Add a virtual memory mapping in a process's page table
 *   from virtual address 'address' to physical page number 'pn'.
 *   The mapping is user-accessible and writeable.
 *
 *****************************************************************************/

int
page_map_user(process_t *proc, int pn, uint32_t address)
{
	pagedirectory_t pgdir = proc->p_pgdir;
	pte_t *pgtable;

	// Find page table
	if (!(pgdir[PDX(address)] & PTE_P)) {
		int pn = page_get_free();
		if (pn < 0)
			panic("No memory available for new page table!");
		page_labels[pn] = MEMLABEL_PAGETABLE;
		pgdir[PDX(address)] = (pn << 12) | PTE_P | PTE_W | PTE_U;
	}
	pgtable = (pte_t *) PTE_ADDR(pgdir[PDX(address)]);

	// It is an error if there's already a page mapped at this address
	if (pgtable[PTX(address)] & PTE_P)
		panic("Attempt to remap address %x!", address);
	
	pgtable[PTX(address)] = (pn * PAGESIZE) | PTE_P | PTE_W | PTE_U;

	// We changed the page table, so we need to invalidate the TLB,
	// which may have cached the old entry.
	invlpg((void *) address);
	
	return 0;
}


/*****************************************************************************
 * pgdir_new
 *
 *   Allocates and returns a new page directory.
 *   The page directory's initial state maps all kernel addresses
 *   (that is, all addresses below 0x100000)
 *   to the corresponding physical addresses.
 *   These mappings are accessible only to the kernel, except that the
 *   console is user-accessible as well.
 *   Thus, the initial page map function works like this:
 *
 *   Virtual Address       Physical Address and Mapping Properties
 *   (Input)               (Output)        
 *   ---------------       ---------------------------------------
 *      0x00000000   ==>   0x00000000   Kernel-Only, Read/Write
 *      0x00001000   ==>   0x00001000   Kernel-Only, Read/Write
 *      0x00002000   ==>   0x00002000   Kernel-Only, Read/Write
 *      ...
 *      0x000B7000   ==>   0x000B7000   Kernel-Only, Read/Write
 *      0x000B8000   ==>   0x000B8000   User or Kernel, Read/Write  (CONSOLE)
 *      0x000B9000   ==>   0x000B9000   Kernel-Only, Read/Write
 *      ...
 *      0x000FE000   ==>   0x000FE000   Kernel-Only, Read/Write
 *      0x000FF000   ==>   0x000FF000   Kernel-Only, Read/Write
 *      0x00100000   ==>   PAGE FAULT
 *      0x00110000   ==>   PAGE FAULT
 *      0x00120000   ==>   PAGE FAULT
 *      ...
 *      0xFFFE0000   ==>   PAGE FAULT
 *      0xFFFF0000   ==>   PAGE FAULT
 *
 *****************************************************************************/

pagedirectory_t
pgdir_new(void)
{
	int pgdir_pn, pgtbl_pn, i;
	pte_t *pgdir, *pgtbl;
	
	pgdir_pn = page_get_free();
	page_labels[pgdir_pn] = MEMLABEL_PAGETABLE;
	pgtbl_pn = page_get_free();
	page_labels[pgtbl_pn] = MEMLABEL_PAGETABLE;
	pgdir = (pte_t *) (pgdir_pn * PAGESIZE);
	pgtbl = (pte_t *) (pgtbl_pn * PAGESIZE);

	// page directory points to page table
	memset(pgdir, 0, PAGESIZE);
	pgdir[0] = (uint32_t) pgtbl | PTE_P | PTE_W | PTE_U;

	// set page table mappings: everything defaults to off
	memset(pgtbl, 0, PAGESIZE);
	// kernel is accessible, but not by user processes
	for (i = 0; i < PAGENUM(PROCESS_START); i++)
		pgtbl[i] = (i * PAGESIZE) | PTE_P | PTE_W;
	// ... except the console is accessible by user processes
	for (i = (uint32_t) CONSOLE_BEGIN;
	     i < (uint32_t) CONSOLE_END;
	     i += PAGESIZE)
		pgtbl[PTX(i)] |= PTE_U;

	return pgdir;
}


/*****************************************************************************
 * MEMORY LABELS FUNCTIONS
 *
 * memory_labels_init
 *   Initializes memory labels.
 *
 * show_physical_memory
 * show_virtual_memory(proc)
 *   Shows the physical or virtual address space.
 *
 * show_virtual_memory_flipper
 *   Shows a process's virtual address space.  Repeated calls will gradually
 *   flip through all four processes.
 *
 *****************************************************************************/

void
memory_labels_init(void)
{
	int p, label;
	extern char _end;

	for (p = 0; p < MAX_PHYSICAL_MEMORY; p += PAGESIZE) {
		if (p == 0 || (p >= IOPHYSMEM && p < EXTPHYSMEM))
			label = MEMLABEL_RESERVED;
		else if ((p >= KERNEL_START && p < (uint32_t) &_end)
			 || p == KERNEL_STACK_TOP - PAGESIZE)
			label = MEMLABEL_KERNEL;
		else
			label = MEMLABEL_FREE;
		page_labels[PAGENUM(p)] = label;
	}
}

static const uint16_t memlabel_colors[] = {
	'.' | 0x0700, 'R' | 0x0700, 'P' | 0x0B00, 'K' | 0x0D00,
	'1' | 0x0C00, '2' | 0x0A00, '3' | 0x0900, '4' | 0x0E00,
	'.' | 0x7000, 'R' | 0xF000, 'P' | 0xB000, 'K' | 0xD000,
	'1' | 0xC000, '2' | 0xA000, '3' | 0x9000, '4' | 0xE000,
	' ' | 0x0000
};

void
show_physical_memory(void)
{
	uint16_t *cptr;
	int pn;

	console_printf(0, 32, "%CPHYSICAL MEMORY", 0x0F00);
	cptr = (uint16_t *) CONSOLE_BEGIN + 80 + 12;
	for (pn = 0; pn < PAGENUM(0x200000); pn++) {
		if (pn % 64 == 0)
			console_printf(1 + pn / 64, 3, "%08X ", pn << 12);
		*cptr++ = memlabel_colors[page_labels[pn]];
		if (pn % 64 == 63)
			cptr += 16;
	}
}

void
show_virtual_memory(process_t *proc)
{
	uint16_t *cptr, color;
	pagedirectory_t pgdir = proc->p_pgdir;
	pte_t *pgtbl;
	int pn;

	console_printf(10, 26, "%CVIRTUAL ADDRESS SPACE FOR %c", 0x0F00,
		       (proc ? '1' + proc->p_pid : 'K'));
	cptr = (uint16_t *) CONSOLE_BEGIN + 80*11 + 12;
	pgtbl = (pte_t *) PTE_ADDR(pgdir[0]);
	for (pn = 0; pn < PAGENUM(0x300000); pn++) {
		pte_t pte = pgtbl[pn];
		if (!(pte & PTE_P))
			color = 16;
		else if (!(pte & PTE_U))
			color = page_labels[PTE_ADDR(pte) / PAGESIZE];
		else
			color = page_labels[PTE_ADDR(pte) / PAGESIZE] + 8;
		if (pn % 64 == 0)
			console_printf(11 + pn / 64, 3, "%08X ", pn << 12);
		*cptr++ = memlabel_colors[color];
		if (pn % 64 == 63)
			cptr += 16;
	}
}

void
show_virtual_memory_flipper(void)
{
	static int call_count = 0;
	static int flipper_pid = 0;

	if (++call_count >= MEMORY_FLIPPER_SPEED) {
		call_count = 0;
		flipper_pid = (flipper_pid + 1) % NPROCESSES;
	}
	
	show_virtual_memory(&processes[flipper_pid]);
}
